/* -*- mode: c++; c-basic-offset: 3; -*- */
// STREAMING_CONDITIONER Resample, high-pass filter and whiten time series data
//
// STREAMING_CONDITIONER resamples, high-pass filters and whitens time series
// data prior to analysis by the Q transform.  The data is first resampled to
// desired rate (specified in the tiling structure). Then, a zero-phase high-
// pass filter is applied at the minimum analysis frequency specified in the
// tiling structure.  The resulting filtered data are whitened at a frequency
// resolution equal to the minimum analysis bandwidth(?). requested in
// the tiling structure.  Note that the resulting whitened data is returned as
// a frequency series, not a time series.  In addition, the returned frequency
// series extends from zero frequency to the Nyquist frequency.  As a result,
// they are of length N / 2 + 1, where N is the length of the individual input
// time series.
//
// To enable recovery of calibrated amplitude and phase information, WCONDITION
// also returns the effective frequency domain coefficients of the combined high
// pass and whitening filters for each channel.
//
// WCONDITION also permits double whitening of the data to support true matched
// filtering.  Regardless of whether double whitening is requested or not, the
// returned frequency-domain filter coefficients always correspond to the single
// whitened case.
//
// usage:
//
//   streaming_conditioner(rawData, tiling, doubleWhiten);
//
//   rawData               vector of input time series
//   tiling                discrete Q transform tiling structure from WTILE
//   doubleWhiten          bool for double whitening
//
//   conditionedData       cell array of conditioned output frequency series
//   coefficients          cell array of frequency domain filter coefficients
//   doubleWhiten          boolean flag to enable double whitening
//
// There is also an alternative output syntax, which provides access to the
// intermediate raw and high pass filtered data for use by WSCAN.
//
//   [rawData, highPassedData, whitenedData, coefficients] = 
//       streaming_conditioner(rawData, tiling, doubleWhiten);
//
//   rawData               cell array of unconditioned frequency series
//   highPassedData        cell array of high pass filtered frequency series
//   whitenedData          cell array of whitened frequency series data
//   coefficients          cell array of frequency domain filter coefficients
//
// See also WTILE, WTRANSFORM, WTHRESHOLD, WSELECT, WSEARCH, WEVENT, WSCAN,
// and WEXAMPLE.
// John Zweizig <john.zweizig@ligo.org>
// Based on algoritm by: Shourov K. Chatterji <shourov@ligo.mit.edu>


#include "wtypes.hh"
#include "MeanMedianPSD.hh"
#include "autopipe.hh"
#include "resampler.hh"
#include <stdexcept>

class IIRFilter;

namespace wpipe {

   class wtile;

   /**  Condition the data from one input channel. The conditioning process
     *  high-passes and whitenes the data. It then returns one or all of the
     *  resulting time series or DFTs. These results include:
     *  * A Raw data series.
     *  * A high-passed data time series
     *  * A whitened time series
     *  * A high-passed DFT
     *  * A whitened DFT
     *  * A DFT containing the whitening transfer function.
     */
   class streaming_conditioner {
   public:
      /**  Construct a mullti-channel streaming_conditioner instance for 
        *  the specified number of channels. The channel sample rate, 
	*  filtering, etc. are specified by the wtile instance.
	*  
	*  \param tiling       Tiling specification (wtile)
	*  \param doubleWhiten Doubly whiten the data is set.
	*  \param dbugLevel    Number of channels processed in a group.
        */
      streaming_conditioner(const wtile& tiling, bool doubleWhiten=false,
			    int dbugLevel = 0);
      
      /**  Destroy a streaming_conditioner instance.
       */
      ~streaming_conditioner(void);

      /**  Condition a single channel time series and buffer the output. Valid
        *  data corresponding to the time interval starting at \a t0 with 
	*  duration specified by at initialization by the wtile::duration 
	*  parameter are Fourier transformed and whitened. Whitening is 
	*  controlled by the \a valid argument and a valid (non-zero) \a t0.
	*  A zero \a t0 is user to indicate that the high-pass filter and 
	*  whitener must be initialized and indicates data are to be treated 
	*  as invalid.
	*  \param rawData raw input data
	*  \param t0 start of data segment to be conditioned.
	*  \param valid Data are valid for analysis
	*  \return 0: successful, 1: insufficient data, 2: data starts after t0,
	*         -1: error occurred
        */
      int condition(const TSeries& rawData, const Time& t0, bool valid=true);

      /**  Reference to a vector of DFTs containing the transfer function 
        *  coefficients for each processed channel.
	*/
      const containers::DFT& coefficientDFT(void) const;

      /** Reference to the coefficient DFT of the first/only channel processed.
       */
      void coefficientDFT(containers::DFT& dftvec) const;

      /**  currentTime() returns the time of the next expected raw (input)
        *  data sample.
	*/
      const Time& currentTime() const;

      /**  Get the filter time delay
       */
      Interval getTimeDelay(void) const;
      
      /** Reference to a vector of DFTs of hi-passed data.
       */
      const containers::DFT& highPassedDFT(void) const;
      TSeries highPassedData(void) const;
      
      /** Reference to a vector of DFTs for raw data.
       */
      const containers::DFT& rawDFT(void);

      /**  Reset the filters, etc.
       */
      void reset(void);
      
      /** Time series containing the whitened data from the first/only channel.
       */
      TSeries whitenedData(void) const;

      /**  Copy a vector of time series created by whitening the channels.
       */
      void whitenedData(TSeries& ts) const;

      /** Reference to a DFT containing whitened data of the first/only channel.
       */
      const containers::DFT& whitenedDFT(void) const;

      /** Copy a vector of whitened DFTs to the argument location.
       */
      void whitenedDFT(containers::DFT& dftvec) const;

   private:
      //--------------------------------  resample input data
      double          fSample;
      Interval        duration;
      size_t          dataLength;
      resampler       resample;
      TSeries         raw;

      //--------------------------------  resampled raw data
      containers::DFT rawDft;
      containers::DFT coefficients;

      //--------------------------------  High-pass data
      double          highPassCutoff;
      auto_pipe       highPassFilter;
      containers::fSeries hpfResponse;
      TSeries         highPassData;
      containers::DFT highPassDft;

      //--------------------------------  Whitener
      enum {wNone, wSingle, wDouble} whiten_stat;
      MeanMedianPSD   whitener;
      containers::DFT whitenedDft;

      //--------------------------------  Miscellaneous parameters
      Time            current;
      int             debugLevel;
   };

   inline const containers::DFT& 
   streaming_conditioner::coefficientDFT(void) const {
      return coefficients;
   }
   
   inline const Time& 
   streaming_conditioner::currentTime(void) const {
      return current;
   }

   inline const containers::DFT&
   streaming_conditioner::highPassedDFT(void) const {
      return highPassDft;
   }

   inline const containers::DFT&
   streaming_conditioner::whitenedDFT(void) const {
      return whitenedDft;
   }

} // namespace wpipe
